import os
from datetime import datetime, timedelta
from typing import overload

import aiofiles
from nonebot.adapters.onebot.v11 import MessageEvent
from nonebot_plugin_orm import AsyncSession, get_session
from pydantic import BaseModel, Field
from sqlalchemy import delete, select
from typing_extensions import Self

from .config import get_cache_dir, get_config
from .models import THEME_TYPE, OmikujiData
from .sql_models import OmikujiCache as SQLOmikujiCache
from .sql_models import db_lock


class OmikujiCache(BaseModel):
    data: OmikujiData
    timestamp: datetime = Field(default_factory=datetime.now)


async def cache_omikuji(event: MessageEvent, data: OmikujiData) -> None:
    CACHE_DIR = get_cache_dir()
    os.makedirs(CACHE_DIR, exist_ok=True)
    cache_file = CACHE_DIR / f"{event.user_id!s}.json"
    cache = OmikujiCache(data=data)
    async with aiofiles.open(cache_file, "w", encoding="utf-8") as f:
        await f.write(cache.model_dump_json())


async def get_cached_omikuji(event: MessageEvent) -> OmikujiData | None:
    CACHE_DIR = get_cache_dir()
    os.makedirs(CACHE_DIR, exist_ok=True)
    cache_file = CACHE_DIR / f"{event.user_id!s}.json"
    if not cache_file.exists():
        return None
    async with aiofiles.open(cache_file, encoding="utf-8") as f:
        cache = OmikujiCache.model_validate_json(await f.read())
    if cache.timestamp.date() == datetime.now().date():
        return cache.data
    else:
        os.remove(str(cache_file))


class OmikujiCacheContent(BaseModel):
    content: str


class OmikujiCacheData(BaseModel):
    level: str
    theme: THEME_TYPE
    sections: dict[str, list[str]]
    intro: list[OmikujiCacheContent]
    maxim: list[OmikujiCacheContent]
    end: list[OmikujiCacheContent]
    divine_title: list[OmikujiCacheContent]
    sign_number: list[OmikujiCacheContent]
    created_date: str
    updated_date: str

    @overload
    @classmethod
    async def get(cls, level: str, theme: THEME_TYPE) -> Self | None: ...

    @overload
    @classmethod
    async def get(cls, level: str) -> dict[str, Self]: ...

    @classmethod
    async def get(
        cls, level: str, theme: THEME_TYPE | None = None
    ) -> Self | dict[str, Self] | None:
        async with db_lock(theme, level):
            async with get_session() as session:
                await cls._expire_cache(session=session)
                if theme:
                    stmt = select(SQLOmikujiCache).where(
                        SQLOmikujiCache.level == level,
                        SQLOmikujiCache.theme == theme,
                    )
                    result = await session.execute(stmt)
                    data = result.scalar_one_or_none()
                    if not data:
                        return None
                    return cls.model_validate(data, from_attributes=True)
                else:
                    stmt = select(SQLOmikujiCache).where(
                        SQLOmikujiCache.level == level,
                    )
                    result = await session.execute(stmt)
                    data = result.scalars().all()
                    return {
                        model.theme: cls.model_validate(model, from_attributes=True)
                        for model in data
                    }

    @classmethod
    async def cache_omikuji(cls, data: OmikujiData) -> None:
        config = get_config()

        def unique_append(ls: list[dict[str, str]], content: str):
            d = OmikujiCacheContent(content=content).model_dump()
            if d not in ls:
                ls.append(d)
            while len(d) > config.omikuji_long_cache_update_max_count:
                ls.pop(0)

        async with db_lock(data.theme, data.level):
            async with get_session() as session:
                await cls._expire_cache(session=session)
                stmt = (
                    select(SQLOmikujiCache)
                    .where(
                        SQLOmikujiCache.theme == data.theme,
                        SQLOmikujiCache.level == data.level,
                    )
                    .with_for_update()
                )
                if (
                    cache := (await session.execute(stmt)).scalar_one_or_none()
                ) is None:
                    cache = SQLOmikujiCache(
                        theme=data.theme,
                        level=data.level,
                        sections={i.name: [i.content] for i in data.sections},
                        intro=[
                            OmikujiCacheContent(content=data.sign_number).model_dump()
                        ],
                        maxim=[OmikujiCacheContent(content=data.maxim).model_dump()],
                        end=[OmikujiCacheContent(content=data.end).model_dump()],
                        divine_title=[
                            OmikujiCacheContent(content=data.divine_title).model_dump()
                        ],
                        sign_number=[
                            OmikujiCacheContent(content=data.sign_number).model_dump()
                        ],
                    )
                    session.add(cache)
                    await session.commit()
                    await session.refresh(cache)

                sections = cache.sections
                intro = cache.intro
                maxim = cache.maxim
                end = cache.end
                divine_title = cache.divine_title
                sign_number = cache.sign_number

                for i in data.sections:
                    if i.name not in sections:
                        sections[i.name] = []
                    if i.content not in sections[i.name]:
                        sections[i.name].append(i.content)

                unique_append(intro, data.intro)
                unique_append(maxim, data.maxim)
                unique_append(end, data.end)
                unique_append(divine_title, data.divine_title)
                unique_append(sign_number, data.sign_number)

                cache.intro = intro
                cache.sections = sections
                cache.maxim = maxim
                cache.end = end
                cache.divine_title = divine_title
                cache.sign_number = sign_number

                await session.commit()

    @staticmethod
    async def _expire_cache(*, session: AsyncSession) -> None:
        config = get_config()
        if not config.omikuji_long_cache_mode:
            if config.omikuji_cache_expire_days > 0:
                expire_time = datetime.now() - timedelta(
                    days=config.omikuji_cache_expire_days
                )
                await session.execute(
                    delete(SQLOmikujiCache).where(
                        SQLOmikujiCache.created_date < expire_time.strftime("%Y-%m-%d")
                    )
                )
                await session.commit()
            if config.omikuji_cache_update_expire_days > 0:
                expire_time = datetime.now() - timedelta(
                    days=config.omikuji_cache_update_expire_days
                )
                await session.execute(
                    delete(SQLOmikujiCache).where(
                        SQLOmikujiCache.updated_date < expire_time.strftime("%Y-%m-%d")
                    )
                )
                await session.commit()



from __future__ import annotations

from pathlib import Path

from nonebot import get_plugin_config
from nonebot_plugin_localstore import get_plugin_cache_dir
from pydantic import BaseModel, model_validator


class Config(BaseModel):
    """
    Configuration for nonebot_plugin_omikuji
    """

    omikuji_send_by_chat: bool = False  # 是否交给模型进行二次响应
    enable_omikuji: bool = True  # 是否启用
    omikuji_add_system_prompt: bool = (
        True  # 是否加入SuggarChat的系统提示(生成更符合角色设定的答案)
    )
    omikuji_use_cache: bool = True  # 是否使用语料库的缓存（LLM生成后的缓存）
    omikuji_cache_expire_days: int = (
        14  # 御神签语料缓存有效期，创建时间超过该天数之前会被清除（-1表示长期有效）
    )
    omikuji_cache_update_expire_days: int = (
        7  # 更新时间差大于这个数值就会清除缓存(-1表示不检查更新时间)
    )
    omikuji_long_cache_mode: bool = True  # 启用长期缓存模式(不会清除缓存)
    omikuji_long_cache_update: bool = True  # 仅在语料库长期模式下生效，是否自动更新语料
    omikuji_long_cache_update_days: int = 3  # 仅在语料库长期模式下生效，同一个Level和主题添加缓存内容的间隔天数(0为不更新,即使命中内容过少也会命中)
    omikuji_long_cache_update_max_count: int = (
        100  # 仅在语料库长期模式下生效，添加缓存内容的最大数量
    )

    @model_validator(mode="after")
    def check(self) -> Config:
        if self.omikuji_long_cache_update_max_count < 1:
            raise ValueError(
                "omikuji_long_cache_update_max_count must be greater than 0"
            )
        if self.omikuji_long_cache_update_days < 1:
            raise ValueError("omikuji_long_cache_update_days must be greater than 0")
        return self


CONFIG: Config = get_plugin_config(Config)
CACHE_DIR: Path = get_plugin_cache_dir()


def get_config() -> Config:
    return CONFIG


def get_cache_dir() -> Path:
    return CACHE_DIR



from __future__ import annotations

import asyncio
from datetime import datetime
from functools import lru_cache

from nonebot_plugin_orm import Model
from sqlalchemy import JSON, Index, String, UniqueConstraint
from sqlalchemy.orm import Mapped, mapped_column

from .models import THEME_TYPE


@lru_cache(1024)
def db_lock(*args, **kwargs):
    return asyncio.Lock()


class OmikujiCache(Model):
    __tablename__ = "omikuji_cache"
    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    level: Mapped[str] = mapped_column(String(64), nullable=False)
    theme: Mapped[THEME_TYPE] = mapped_column(String(64), nullable=False)
    sections: Mapped[dict[str, list[str]]] = mapped_column(
        JSON,
        default={},
        nullable=False,
    )  # {"板块":["板块内容"]}
    intro: Mapped[list[dict[str, str]]] = mapped_column(
        JSON,
        default=[],
        nullable=False,
    )  # [{"content":"内容"}]
    maxim: Mapped[list[dict[str, str]]] = mapped_column(
        JSON,
        default=[],
        nullable=False,
    )
    end: Mapped[list[dict[str, str]]] = mapped_column(
        JSON,
        default=[],
        nullable=False,
    )
    divine_title: Mapped[list[dict[str, str]]] = mapped_column(
        JSON,
        default=[],
        nullable=False,
    )
    sign_number: Mapped[list[dict[str, str]]] = mapped_column(
        JSON,
        default=[],
        nullable=False,
    )
    created_date: Mapped[str] = mapped_column(
        String(30), default=lambda: datetime.now().strftime("%Y-%m-%d"), nullable=False
    )
    updated_date: Mapped[str] = mapped_column(
        String(30),
        default=lambda: datetime.now().strftime("%Y-%m-%d"),
        onupdate=lambda: datetime.now().strftime("%Y-%m-%d"),
        nullable=False,
    )
    __table_args__ = (
        Index("ix_omikuji_cache_level_theme", "level", "theme"),
        UniqueConstraint("level", "theme", name="uq_omikuji_cache_level_theme"),
    )



from __future__ import annotations

import random
from typing import Literal

from nonebot_plugin_suggarchat.API import (
    FunctionDefinitionSchema,
    FunctionParametersSchema,
    FunctionPropertySchema,
    ToolFunctionSchema,
)
from pydantic import BaseModel, Field

LEVEL = ["大吉", "吉", "中吉", "小吉", "末吉", "凶", "大凶"]
OMIKUJI_THEMES = [
    "综合运势",  # 默认选项
    "恋爱姻缘",  # 情感
    "学业考试",  # 学习
    "事业财运",  # 工作
    "健康平安",  # 健康
    "人际和谐",  # 关系
    "旅行出行",  # 出行
    "樱花时节",  # 季节/文化
    "星幽秘境",  # 奇幻
    "灵感创意",  # 创作
]
THEME_TYPE = Literal[
    "综合运势",
    "恋爱姻缘",
    "学业考试",
    "事业财运",
    "健康平安",
    "人际和谐",
    "旅行出行",
    "樱花时节",
    "星幽秘境",
    "灵感创意",
]


def random_level() -> str:
    return random.choice(LEVEL)


class OmikujiSections(BaseModel):
    name: str
    content: str


class OmikujiData(BaseModel):
    level: str = Field(default_factory=random_level)
    theme: THEME_TYPE
    sign_number: str
    divine_title: str
    sections: list[OmikujiSections]
    maxim: str
    intro: str
    end: str


OMIKUJI_SCHEMA_META = ToolFunctionSchema(
    strict=True,
    type="function",
    function=FunctionDefinitionSchema(
        name="omikuji_content",
        description="获取御神签签文",
        parameters=FunctionParametersSchema(
            type="object",
            properties={
                "theme": FunctionPropertySchema(
                    type="string",
                    description="御神签主题，必须严格遵守给出的枚举值",
                    enum=OMIKUJI_THEMES,
                ),
                "sign_number": FunctionPropertySchema(
                    type="string",
                    description="御神签编号(随机中文大写数字)",
                ),
                "divine_title": FunctionPropertySchema(
                    type="string",
                    description="神签标题/天启名（2-4字奇幻风格名称）",
                ),
                "sections": FunctionPropertySchema(
                    type="array",
                    items=FunctionPropertySchema(
                        type="object",
                        properties={
                            "name": FunctionPropertySchema(
                                type="string", description="分类名称"
                            ),
                            "content": FunctionPropertySchema(
                                type="string", description="分类的预言内容（1～2句话）"
                            ),
                        },
                        description="签文主体",
                        required=["name", "content"],
                    ),
                    description="签文主体",
                    maxItems=8,
                    minItems=4,
                    uniqueItems=True,
                ),
                "maxim": FunctionPropertySchema(
                    type="string", description="一句箴言/和歌（结尾注明出处）"
                ),
                "intro": FunctionPropertySchema(
                    type="string",
                    description="主题引入(不包含引号)：e.g. '「欢迎来到古树根下的祠堂。异界之风正为你捎来命运的启示…」\n'",
                ),
                "end": FunctionPropertySchema(
                    type="string",
                    description="主题总结(不包含引号)",
                ),
            },
            required=[
                "theme",
                "sign_number",
                "divine_title",
                "sections",
                "maxim",
                "intro",
            ],
        ),
    ),
)

FUNC_META = ToolFunctionSchema(
    strict=True,
    function=FunctionDefinitionSchema(
        name="omikuji",
        description="抽取一个御神签",
        parameters=FunctionParametersSchema(
            type="object",
            properties={
                "theme": FunctionPropertySchema(
                    type="string",
                    description="御神签主题（如果包含不良内容则随机选择）",
                )
            },
            required=["theme"],
        ),
    ),
)



import contextlib
import os
from datetime import datetime

from nonebot import get_driver, logger
from nonebot.plugin import PluginMetadata, require

require("nonebot_plugin_suggarchat")
require("nonebot_plugin_localstore")
require("nonebot_plugin_orm")
from importlib import metadata

from nonebot_plugin_suggarchat.API import Menu, ToolsManager

from . import commands, llm_tool, sql_models
from .cache import OmikujiCache
from .config import get_cache_dir, get_config
from .llm_tool import TOOL_DATA

__plugin_meta__ = PluginMetadata(
    name="御神签",
    description="依赖SuggarChat的聊天御神签抽签插件模块",
    usage="/omikuji [板块]\n/omikuji 解签\n或者使用聊天直接抽签。",
    type="application",
    homepage="https://github.com/LiteSuggarDEV/nonebot_plugin_omikuji",
    supported_adapters={"~onebot.v11"},
)

__all__ = ["commands", "llm_tool", "sql_models"]


@get_driver().on_startup
async def init():
    version = "Unknown"
    with contextlib.suppress(Exception):
        version = metadata.version("nonebot_plugin_omikuji")
        if "dev" in version:
            logger.warning("当前版本为开发版本，可能存在不稳定情况！")
    logger.info(f"Loading OMIKUJI V{version}......")
    conf = get_config()
    if conf.enable_omikuji:
        ToolsManager().register_tool(TOOL_DATA)
        Menu().reg_menu("omikuji", "抽御神签", "[可选]主题")
    logger.info("正在初始化缓存数据......")
    os.makedirs(get_cache_dir(), exist_ok=True)
    for cache in get_cache_dir().glob("*.json"):
        if cache is not None:
            with cache.open("r", encoding="utf-8") as f:
                data = OmikujiCache.model_validate_json(f.read())
            if not data.timestamp.date() == datetime.now().date():
                os.remove(str(cache))
    logger.info("缓存数据初始化完成！")



import typing

from nonebot import get_bot, logger
from nonebot.adapters.onebot.v11 import MessageEvent
from nonebot_plugin_suggarchat.API import (
    ToolContext,
    ToolData,
)

from nonebot_plugin_omikuji.cache import cache_omikuji, get_cached_omikuji

from .config import get_config
from .models import FUNC_META, OmikujiData
from .utils import generate_omikuji

LEVEL = ["大吉", "吉", "中吉", "小吉", "末吉", "凶", "大凶"]


def format_omikuji(data: OmikujiData, user_name: str | None = ""):
    ln = "\n"
    msg = f"""{data.intro}
{(user_name + "，" if user_name else "")}你的签上刻了什么？

＝＝＝ 御神签 第{data.sign_number} ＝＝＝
✨ 天启：{data.divine_title}
🌸 运势：{data.level} - {data.theme}

{"".join(f"▫ {section.name}{ln}{section.content}{ln}" for section in data.sections)}

⚖ 真言偈：{data.maxim}

{data.end}
"""
    return msg


async def omikuji(ctx: ToolContext):
    logger.info("获取御神签")
    nb_event: MessageEvent = typing.cast(MessageEvent, ctx.event.get_nonebot_event())
    is_group = hasattr(nb_event, "group_id")
    bot = get_bot(str(ctx.event._nbevent.self_id))

    if (data := await get_cached_omikuji(nb_event)) is None:
        await bot.send(
            ctx.event._nbevent,
            "轻轻摇动古老的签筒，竹签哗啦作响... 心中默念所求之事... 一支签缓缓落下。",
        )
        data = await generate_omikuji(ctx.data["theme"], is_group)
        await cache_omikuji(nb_event, data)
    if get_config().omikuji_send_by_chat:
        return data.model_dump_json()
    msg = format_omikuji(data)
    await bot.send(nb_event, msg)
    ctx.matcher.cancel_nonebot_process()


TOOL_DATA = ToolData(data=FUNC_META, func=omikuji, custom_run=True)



import random
import typing

from nonebot import on_command
from nonebot.adapters.onebot.v11 import Bot, GroupMessageEvent, Message, MessageEvent
from nonebot.params import CommandArg

from .cache import cache_omikuji, get_cached_omikuji
from .config import get_config
from .models import OMIKUJI_THEMES, THEME_TYPE
from .utils import format_omikuji, generate_omikuji

omikuji = on_command(
    "omikuji",
    aliases={"御神签", "抽签"},
    priority=10,
    block=True,
    rule=lambda: get_config().enable_omikuji,
)


@omikuji.handle()
async def _(bot: Bot, event: MessageEvent, args: Message = CommandArg()):
    if args:
        theme = args.extract_plain_text()
        if theme not in OMIKUJI_THEMES:
            await omikuji.finish(
                f"当前可用御神签主题：{''.join(i + ',' for i in OMIKUJI_THEMES)}"
            )
    else:
        theme = random.choice(list(OMIKUJI_THEMES))
    theme = typing.cast(THEME_TYPE, theme)
    is_group = isinstance(event, GroupMessageEvent)
    if (data := await get_cached_omikuji(event)) is None:
        data = await generate_omikuji(theme, is_group)
        await cache_omikuji(event, data)
    msg = format_omikuji(data)
    await omikuji.finish(msg)



import json
import random
from copy import deepcopy
from datetime import datetime, timedelta

from nonebot import logger
from nonebot_plugin_suggarchat.API import (
    config_manager,
    tools_caller,
)
from nonebot_plugin_suggarchat.utils.models import Message

from .cache import OmikujiCacheData
from .config import get_config
from .models import (
    OMIKUJI_SCHEMA_META,
    THEME_TYPE,
    OmikujiData,
    OmikujiSections,
    random_level,
)


async def _hit_cache_omikuji(
    theme: THEME_TYPE,
    level: str = "",
) -> OmikujiData | None:
    if cache := await OmikujiCacheData.get(level, theme):
        if cache.updated_date < (
            datetime.now() - timedelta(days=get_config().omikuji_cache_expire_days)
        ).strftime("%Y-%m-%d"):
            logger.debug(f"{theme}/{level} cache expired!")
            return
        logger.debug(f"{theme}/{level} cache hit!")
        keys = list(cache.sections.keys())
        random.shuffle(keys)
        sections = [
            OmikujiSections(name=k, content=random.choice(cache.sections[k]))
            for k in keys
        ]
        if len(sections) < 4:
            return
        while len(sections) > 8:
            sections.pop()

        model = OmikujiData(
            level=level,
            theme=theme,
            sections=sections,
            sign_number=random.choice([i.content for i in cache.sign_number]),
            intro=random.choice([i.content for i in cache.intro]),
            divine_title=random.choice([i.content for i in cache.divine_title]),
            maxim=random.choice([i.content for i in cache.maxim]),
            end=random.choice([i.content for i in cache.end]),
        )
        return model


async def generate_omikuji(
    theme: THEME_TYPE,
    is_group: bool = False,
    level: str = "",
) -> OmikujiData:
    config = get_config()
    level = level or random_level()
    if config.omikuji_use_cache:
        if cache := await _hit_cache_omikuji(theme, level):
            return cache
    logger.debug(f"theme: {theme}, level: {level} Cache miss")
    system_prompt = Message.model_validate(
        deepcopy(
            config_manager.group_train if is_group else config_manager.private_train
        )
    )
    assert isinstance(system_prompt.content, str)
    system_prompt.content += "\n你现在需要结合你的角色设定生成御神签。"
    user_prompt = Message(
        role="user",
        content=f"御神签的运势是：'{level}'\n现在生成一张主题为：'{theme}'的御神签",
    )
    msg_input = [system_prompt, user_prompt]
    data = await tools_caller(
        messages=msg_input, tools=[OMIKUJI_SCHEMA_META], tool_choice="required"
    )
    assert data.tool_calls
    args = json.loads(data.tool_calls[0].function.arguments)
    args["level"] = level
    args["theme"] = theme
    model = OmikujiData.model_validate(args)
    if level:
        model.level = level
    if config.omikuji_use_cache:
        await OmikujiCacheData.cache_omikuji(model)
    return model


def format_omikuji(data: OmikujiData, user_name: str = ""):
    ln = "\n"
    msg = f"""{data.intro}
{(user_name + "，" if user_name else "")}你的签上刻了什么？

＝＝＝ 御神签 第{data.sign_number} ＝＝＝
✨ 天启：{data.divine_title}
🌸 运势：{data.level} - {data.theme}

{"".join(f"▫ {section.name}{ln}{section.content}{ln}" for section in data.sections)}

⚖ 真言偈：{data.maxim}

{data.end}
"""
    return msg
